import asyncio
from functools import wraps, _make_key
from .moka_py import Moka


__all__ = ["Moka", "cached"]


def cached(maxsize=128, typed=False, *, ttl=None, tti=None):
    cache = Moka(maxsize, ttl, tti)

    def dec(fn):
        if asyncio.iscoroutinefunction(fn):
            @wraps(fn)
            async def inner(*args, **kwargs):
                key = _make_key(args, kwargs, typed)
                maybe_value = cache.get(key)
                if maybe_value is not None:
                    return maybe_value
                value = await fn(*args, **kwargs)
                cache.set(key, value)
                return value
        else:
            @wraps(fn)
            def inner(*args, **kwargs):
                key = _make_key(args, kwargs, typed)
                maybe_value = cache.get(key)
                if maybe_value is not None:
                    return maybe_value
                value = fn(*args, **kwargs)
                cache.set(key, value)
                return value

        return inner

    return dec



import asyncio

import pytest

import moka_py
import random


def test_decorator_sync():
    lst = []

    @moka_py.cached()
    def f(x: int, y: int) -> float:
        lst.append((x, y))
        return x / y

    cases = [
        (1, 2), (3, 4), (5, 6), (7, 8)
    ]

    answers = {}

    for _ in range(100):
        copy = cases[:]
        random.shuffle(copy)
        for case in copy:
            result = f(*case)
            if case not in answers:
                answers[case] = result
            else:
                assert answers[case] == result

    # A call with the same arguments occurs only once
    assert len(lst) == len(cases)


@pytest.mark.asyncio
async def test_decorator_async():
    lst = []

    @moka_py.cached()
    async def f(x: int, y: int) -> float:
        lst.append((x, y))
        await asyncio.sleep(0.01)
        return x / y

    cases = [
        (1, 2), (3, 4), (5, 6), (7, 8)
    ]

    answers = {}

    for _ in range(100):
        copy = cases[:]
        random.shuffle(copy)
        for case in copy:
            result = await f(*case)
            if case not in answers:
                answers[case] = result
            else:
                assert answers[case] == result

    # A call with the same arguments occurs only once
    assert len(lst) == len(cases)
    print(lst, answers)
